/*
  port.c - C implementation of routines defined in portable.h 

  Part of FreeRTOS NIOS2 port - Copyright (C) 2006 Kolja Waschk

  This port is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This port is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with FreeRTOS.org; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
  A special exception to the GPL can be applied should you wish to distribute
  a combined work that includes this port, without being obliged to provide
  the source code for any proprietary components.  See the licensing section 
  of http://www.FreeRTOS.org for full details of how and when the exception
  can be applied.

  ***************************************************************************
  This file has been derived from the MicroBlaze port.c from the
  FreeRTOS.org distribution 4.0.4; Copyright (C) 2003-2006 R. Barry.
  See http://www.FreeRTOS.org for documentation, latest information, license 
  and contact details.  Please ensure to read the configuration and relevant 
  port sections of the online documentation.
  ***************************************************************************
*/

/* Standard includes. */
#include <string.h>
#include <stdlib.h>

/* Scheduler includes. */
#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"

/* NIOS2 specific (also included by system library) */
#include "sys/alt_irq.h"
#include "os/alt_hooks.h"
#include "system.h"
#include <reent.h>
extern void vPortYieldASM(void);
extern void vPortContextRestore(void);
extern void vPortLoadParameters(void *pvParameters);

#ifdef ALT_EXCEPTION_STACK
#error TODO: Support for separated exception stack not yet implemented.
extern void *__alt_exception_stack_pointer;
#ifdef ALT_STACK_CHECK
extern void *__alt_exception_stack_limit;
#endif
#endif

/* Tasks are started with interrupts enabled. */
#define portINITIAL_STATUS ((portSTACK_TYPE) 1)

/* Tasks are started with a critical section nesting of 0 - however prior
to the scheduler being commenced we don't want the critical nesting level
to reach zero, so it is initialised to a high value. */
#define portINITIAL_NESTING_VALUE	( 0xff )

/* Counts the nesting depth of calls to portENTER_CRITICAL().  Each task 
maintains it's own count, so this variable is saved as part of the task
context. */
volatile unsigned portBASE_TYPE uxCriticalNesting = portINITIAL_NESTING_VALUE;

/*-----------------------------------------------------------*/

void vPortInitReent(struct _reent *pREENT)
{
  _REENT_INIT_PTR(pREENT);
  _impure_ptr = pREENT;
}

/*-----------------------------------------------------------*/

/* 
 * Initialise the stack of a task so than when vPortRestoreContext
 * switches to the task, it will start the actual task's code.
 */
portSTACK_TYPE *pxPortInitialiseStack( portSTACK_TYPE *pxTopOfStack, pdTASK_CODE pxCode, void *pvParameters )
{
  portSTACK_TYPE *pxContext;
  
  pxContext = pxTopOfStack - 16;
  
  /* This context is constructed so that it appears as if a switch
     occured during vPortLoadParameters. */

  pxContext[ 0] = 0x8016; /* r16 */
  pxContext[ 1] = 0x8017; /* r17 */
  pxContext[ 2] = 0x8018; /* r18 */
  pxContext[ 3] = 0x8019; /* r19 */
  pxContext[ 4] = 0x8020; /* r20 */
  pxContext[ 5] = 0x8021; /* r21 */
  pxContext[ 6] = 0x8022; /* r22 */
  pxContext[ 7] = 0x8023; /* r23 */
  pxContext[ 8] = 0x80F6; /* fp, will be soon reloaded in vLoadParameters */
  pxContext[ 9] = (portSTACK_TYPE)vPortLoadParameters; /* ra */
  pxContext[10] = (portSTACK_TYPE)portINITIAL_STATUS; /* status */
  pxContext[11] = (portSTACK_TYPE)0; /* uxCriticalNesting */
  pxContext[12] = (portSTACK_TYPE)_impure_ptr;
  
  /* The following is what would have been put on stack of a simple
     function that takes just a void* argument. When "returning" from 
     vPortLoadParameters, pvParameters will be loaded into r4 and thus
     is available to pxCode (the task) like a void* argument. */

  pxContext[13] = (portSTACK_TYPE)pvParameters; /* r4 */
  pxContext[14] = (portSTACK_TYPE)pxContext; /* fp */
  pxContext[15] = (portSTACK_TYPE)pxCode; /* ra for vLoadParameters */

  /* Return a pointer to the top of the stack we have generated so this can
  be stored in the task control block for the task. */
  return pxContext;
}
/*-----------------------------------------------------------*/

portBASE_TYPE xPortStartScheduler( void )
{
  
  /* Kick off the first task. */
  asm volatile (" br vPortContextRestore");

	/* Can't get here as the tasks are now running! */
	return pdFALSE;
}
/*-----------------------------------------------------------*/

void vPortEndScheduler( void )
{
	/* Not implemented. */
}
/*-----------------------------------------------------------*/

/*
 * Manual context switch called by portYIELD.
 */
void vPortYield( void )
{
	/* Perform the context switch in a critical section to assure it is
	not interrupted by the tick ISR.  It is not a problem to do this as
	each task maintains it's own interrupt status. */

	portENTER_CRITICAL();
  vPortYieldASM();
	portEXIT_CRITICAL();
}

/* 
 * Handler for the timer interrupt.
 */
 
volatile int xPortTicked;

void vPortTick()
{
	/* Increment the RTOS tick - this might cause a task to unblock. */
	vTaskIncrementTick();

	/* If we are using the preemptive scheduler then we also need to 
  determine	if this tick should cause a context switch. The actual 
  switch is coded in ALT_OS_INT_EXIT(), pulled in from os/alt_hooks.h */
  
#if ( configUSE_PREEMPTION == 1 )
  xPortTicked = 1;
#endif
}

/*-----------------------------------------------------------*/
/* malloc/free: trust Altera's newlib (code from MemMang/3) */

void *pvPortMalloc( size_t xWantedSize )
{
  void *pvReturn;

  vTaskSuspendAll();
  pvReturn = malloc( xWantedSize );
  xTaskResumeAll();

  return pvReturn;
}

void vPortFree( void *pv )
{
  if( pv != NULL )
  {
    vTaskSuspendAll();
    free( pv );
    xTaskResumeAll();
  }
}

